#include <systemc.h>

SC_MODULE(adpcm_RTL)
{
	// Input Port
	sc_in < sc_uint<16> > len;    // buffer size
	sc_in <short> indata;         // input value from buffer
	sc_in <bool> clk;             // system clock
	sc_in <bool> reset;           // reset for adpcm
	
	// Output Port
	sc_out <char> outdata;        // output values of adpcm
	sc_out <bool> outdata_enable; // 0->1 when a new value is computed and ready
	sc_out <bool> indata_enable;  // 0->1 when a new indata is required for continue
	sc_out <bool> len_enable;     // 0->1 when a new buffer is required for continue
	
	// States List
	enum state_adpcm {
		RESET,PRE1,PRE2,STEP01,STEP0,STEP1, STEP2, STEP3, STEP4, STEP5, STEP6
	};
	
	// Internal Signal
	sc_signal < state_adpcm > state;        // current state
	sc_signal < state_adpcm > next_state;   // next state
	
	// Internal Variable
	signed char c;		
	signed char outputbuffer;		/* place to keep previous 4-bit value */
	
	short val;			        /* Current input sample value */
	
	sc_uint<4> sign;			/* Current adpcm sign bit */
	sc_uint<4> delta;			/* Current adpcm output value */
	sc_uint<18> step;			/* Stepsize */
	sc_uint<18> vpdiff;			/* Current change to valpred */
	sc_uint<16> len_value;              	/* Buffer size in input*/	
	
	int valpred;			        /* Predicted output value */
	int diff;			        /* Difference between val and valprev */
	int index;			        /* Current step change index */
	
	bool bufferstep;		        /* toggle between outputbuffer/output */
	
	// Declare Processes
	void StateEvol();                     // FSM, state implementation
	
	// Constructor
	SC_CTOR (adpcm_RTL) {
		
		SC_METHOD (StateEvol);
		sensitive_pos << clk;
		
	}
	
};  // end SC_MODULE

// FSM for adpcm. Sensitive to state signal
void adpcm_RTL::StateEvol()
{
	
	// Static Array Declaration - READONLY ARRAYS
	
	/* Intel ADPCM step variation table */
	int indexTable[16] = {
		-1, -1, -1, -1, 2, 4, 6, 8,
		-1, -1, -1, -1, 2, 4, 6, 8,
	};
	
	int stepsizeTable[89] = {
		7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
		19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
		50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
		130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
		337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
		876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
		2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
		5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
		15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
	};
	
	switch (state) {
		
		case RESET:   // Initial state, execute after a RESET signal	
			c=0;		
			outputbuffer=0;
			val=0;	
			sign=0;	
			delta=0;	
			step=0;	
			vpdiff=0;	
			len_value=0; 
			valpred=0;
			diff=0;	
			index=0;	
			bufferstep=0;
			
			outdata_enable.write(0);
			indata_enable.write(0);
			len_enable.write(0);
			outdata.write(0);
			
			state=PRE1;
			break;
			
		case PRE1:    // ask for a new len value (buffer size)
		len_enable.write(1);
		
		//NB
		outdata_enable.write(0);
		
		state=PRE2;
		break;
		
		case PRE2:   // wait one clock cycle before reading buffer size
			len_enable.write(0);
			
			state=STEP0;
			break;
			
		case STEP0:  // Read buffer size to compute
			len_value=len.read();
			
			if (len_value == 0)  // wait until a len value != 0 is set
				state=PRE1;
			else{                // start algorithm
				step=stepsizeTable[index];
				bufferstep=1;
				outdata_enable.write(0);
				indata_enable.write(1);  // require a new indata value
				state=STEP01;
			}
			break;
			
		case STEP01:    // wait one clock cycle to read a new indata value
			outdata_enable.write(0);
			indata_enable.write(0);
			
			state=STEP1;
			break;
			
		case STEP1: // STEP1 of algorithm
			val=indata.read();       // read new indata value
			diff=val-valpred;
			sign=(diff<0) ? 8 : 0;
			if ( sign ) 
				diff=(-diff);
			len_enable.write(0);
			//outdata_enable.write(0);
			indata_enable.write(0);
			
			state = STEP2;
			break;
			
		case STEP2: // STEP2 of adpcm algorithm
			
			delta=0;
			vpdiff=(step>>3);
			if (diff>=step){
				delta=4;
				diff-=step;
				vpdiff+=step;
			}
			
			step>>=1;			
			if (diff>=step){
				delta |= 2;
				diff-=step;
				vpdiff+=step;
			}
			
			step>>=1;			
			if (diff>=step){
				delta|=1;
				vpdiff+=step;
			}
			
			state = STEP3;		
			break;
			
		case STEP3:  // STEP3 of adpcm algorithm
			if ( sign )
				valpred-=vpdiff;
			else	
				valpred+=vpdiff;
			
			state = STEP4;
			break;
			
		case STEP4: // STEP4 of adpcm algorithm
			if (valpred > 32767) 
				valpred=32767;
			else if (valpred < -32768) 
				valpred=-32768;			
			
			state = STEP5;
			break;
			
		case STEP5: // STEP5 of adpcm algorithm
			delta|= sign;
			index+=indexTable[delta];
			if (index < 0) index=0;
			if (index > 88) index=88;
			step=stepsizeTable[index];
			
			state = STEP6;
			break;
			
		case STEP6: // STEP6 of adpcm algorithm + "for loop" condition checking
			
			len_value=len_value-1;
			
			if (bufferstep)
				outputbuffer=(delta<<4) & 0xf0;
			else
			{
				c=(delta & 0x0f)| outputbuffer;
				outdata.write(c);     // output new value computed
				outdata_enable.write(1); // notify for new value in output				
			}
			
			bufferstep= !bufferstep;			
			if (len_value==0)
			{
				if ( !bufferstep)
				{
					outdata.write(outputbuffer);  // output new value computed
					outdata_enable.write(1); // notify for new value in output
				}
				state=PRE1;           // A New Buffer is required
			}
			else
			{
				indata_enable.write(1);    // A New indata value is required
				state=STEP01;
			}
			
			break;
	} // end switch
}
