#include <iostream>
#include "systemc.h"

// BINARY ENCODED state machine: drctCntl
// State codes definitions:
#define START_DC 0
#define CHK_DRCT_CNTL 1
#define DRCT_CNTL_WAIT_GNT 2
#define DRCT_CNTL_CHK_LOOP 3
#define DRCT_CNTL_WAIT_RDY 4
#define IDLE_FIN 5
#define IDLE_WAIT_GNT 6
#define IDLE_WAIT_RDY 7

#define TX_DIRECT_CONTROL 0
#define TX_IDLE 5

SC_MODULE(directControl_L)
{
  sc_in<bool>		HCTxPortGnt;
  sc_in<bool>		HCTxPortRdy;
  sc_in<bool>		clk;
  sc_in<bool>		directControlEn;
  sc_in<sc_uint<2> >	directControlLineState;
  sc_in<bool>		rst;

  sc_out<sc_uint<8> >	HCTxPortCntl;
  sc_out<sc_uint<8> >	HCTxPortData;
  sc_out<bool>		HCTxPortReq;
  sc_out<bool>		HCTxPortWEn;

  void drctCntl_NextState();
  void drctCntl_CurrentState();
  void drctCntl_RegOutput();

  sc_signal<sc_uint<8> >	next_HCTxPortCntl;
  sc_signal<sc_uint<8> >	next_HCTxPortData;
  sc_signal<bool>		next_HCTxPortReq;
  sc_signal<bool>		next_HCTxPortWEn;

  sc_signal<sc_uint<3> >	CurrState_drctCntl;
  sc_signal<sc_uint<3> >	NextState_drctCntl;
	
  SC_CTOR(directControl_L)
  {
    SC_METHOD(drctCntl_NextState);
    sensitive << directControlLineState << directControlEn << HCTxPortGnt << HCTxPortRdy << HCTxPortReq << HCTxPortWEn << HCTxPortData << HCTxPortCntl << CurrState_drctCntl;

    SC_METHOD(drctCntl_CurrentState);
    sensitive << clk.pos();

    SC_METHOD(drctCntl_RegOutput);
    sensitive << clk.pos();
  }
};

//--------------------------------------------------------------------
// Machine: drctCntl
//--------------------------------------------------------------------
//----------------------------------
// Next State Logic (combinatorial)
//----------------------------------
void directControl_L::drctCntl_NextState()
{
  NextState_drctCntl.write(CurrState_drctCntl.read());
  // Set default values for outputs and signals
  next_HCTxPortReq.write(HCTxPortReq.read());
  next_HCTxPortWEn.write(HCTxPortWEn.read());
  next_HCTxPortData.write(HCTxPortData.read());
  next_HCTxPortCntl.write(HCTxPortCntl.read());

  switch (CurrState_drctCntl.read())
  {
    case START_DC:
      NextState_drctCntl.write(CHK_DRCT_CNTL);
      break;
    case CHK_DRCT_CNTL:
      if (directControlEn.read() == true)	
      {
        NextState_drctCntl.write(DRCT_CNTL_WAIT_GNT);
        next_HCTxPortReq.write(true);
      }
      else
      {
        NextState_drctCntl.write(IDLE_WAIT_GNT);
        next_HCTxPortReq.write(true);
      }
      break;
    case DRCT_CNTL_WAIT_GNT:
      if (HCTxPortGnt.read() == true)	
      {
        NextState_drctCntl.write(DRCT_CNTL_WAIT_RDY);
      }
      break;
    case DRCT_CNTL_CHK_LOOP:
      next_HCTxPortWEn.write(false);
      if (directControlEn.read() == false)	
      {
        NextState_drctCntl.write(CHK_DRCT_CNTL);
        next_HCTxPortReq.write(false);
      }
      else
      {
        NextState_drctCntl.write(DRCT_CNTL_WAIT_RDY);
      }
      break;
    case DRCT_CNTL_WAIT_RDY:
      if (HCTxPortRdy.read() == true)	
      {
        NextState_drctCntl.write(DRCT_CNTL_CHK_LOOP);
        next_HCTxPortWEn.write(true);
//         next_HCTxPortData <= {6'b000000, directControlLineState};
        next_HCTxPortData.write((0x00)|(directControlLineState.read()));
        next_HCTxPortCntl.write(TX_DIRECT_CONTROL);
      }
      break;
    case IDLE_FIN:
      next_HCTxPortWEn.write(false);
      next_HCTxPortReq.write(false);
      NextState_drctCntl.write(CHK_DRCT_CNTL);
      break;
    case IDLE_WAIT_GNT:
      if (HCTxPortGnt.read() == true)	
        NextState_drctCntl.write(IDLE_WAIT_RDY);
      break;
    case IDLE_WAIT_RDY:
      if (HCTxPortRdy.read() == true)	
      {
        NextState_drctCntl.write(IDLE_FIN);
        next_HCTxPortWEn.write(true);
        next_HCTxPortData.write(0x00);
        next_HCTxPortCntl.write(TX_IDLE);
      }
      break;
  }
}

//----------------------------------
// Current State Logic (sequential)
//----------------------------------
void directControl_L::drctCntl_CurrentState()
{
  if (rst.read())	
    CurrState_drctCntl.write(START_DC);
  else
    CurrState_drctCntl.write(NextState_drctCntl.read());
}

//----------------------------------
// Registered outputs logic
//----------------------------------
void directControl_L::drctCntl_RegOutput()
{
  if (rst.read())	
  {
    HCTxPortCntl.write(0x00);
    HCTxPortData.write(0x00);
    HCTxPortWEn.write(false);
    HCTxPortReq.write(false);
  }
  else
  {
    HCTxPortCntl.write(next_HCTxPortCntl.read());
    HCTxPortData.write(next_HCTxPortData.read());
    HCTxPortWEn.write(next_HCTxPortWEn.read());
    HCTxPortReq.write(next_HCTxPortReq.read());
  }
}

